from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioProxiesBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_proxies'
    class_config = 'proxies-current-resource'
    template = 'proxies'



from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder


def ensure_parent_modules(builder: ModelBuilder, path: Path,
                          *,
                          ends_at: str = None, max_depth=5):
    depth = 0
    path = path.parent
    # 1st sanity check - maximum depth, path must not be a UNC drive name
    while path and path.name and depth < max_depth:
        depth += 1
        # 2nd sanity check - the path must not contain .git
        if builder.filesystem.exists(path.joinpath('.git')):
            break

        init_py_path = path.joinpath('__init__.py')

        # get python output so that other can write into it if required
        # and there are no disk-level conflicts
        builder.get_output('python', init_py_path)

        if path.name == ends_at:
            break
        path = path.parent



import copy
import pathlib
from typing import Dict, Callable
from pathlib import Path

import munch
from jsonpointer import resolve_pointer

from .exceptions import IncludedFileNotFoundException
from .utils.deepmerge import deepmerge
from .utils.hyphen_munch import HyphenMunch


class ModelSchema:
    OAREPO_USE = 'oarepo:use'

    def __init__(self, file_path, content=None,
                 included_models: Dict[str, Callable] = None,
                 loaders=None):
        """
        Creates and parses model schema

        :param file_path: path on the filesystem to the model schema file
        :param content:   if set, use this content, otherwise load the file_path
        :param included_models: a dictionary of file_id to callable that returns included json.
                The callable expects a single parameter, an instance of this schema
        """

        self.file_path = file_path
        self.included_schemas = included_models or {}
        self.loaders = loaders

        if content is not None:
            self.schema = content
        else:
            self.schema = copy.deepcopy(self._load(file_path))

        self._resolve_references(self.schema, [])

        self.schema.setdefault('settings', {})
        self.schema['settings'].setdefault('plugins', {})
        self.schema = munch.munchify(self.schema, factory=HyphenMunch)

    def get(self, key):
        return self.schema.get(key, None)

    def set(self, key, value):
        self.schema[key] = value

    @property
    def settings(self):
        return self.schema.settings

    def merge(self, another):
        self.schema = munch.munchify(deepmerge(another, self.schema, []), factory=HyphenMunch)

    def _load(self, file_path, content=None):
        """
        Loads a json/json5 file on the path

        :param file_path: file path on filesystem
        :return: parsed json
        """
        extension = pathlib.Path(file_path).suffix.lower()[1:]
        if extension in self.loaders:
            return self.loaders[extension](file_path, self, content=content)

        raise Exception(f'Can not load {file_path} - no loader has been found for extension {extension} '
                        f'in entry point group oarepo_model_builder.loaders')

    def _load_included_file(self, location):
        """
        Resolve and load an included file. Internal method called when loading schema.
        If the included file contains a json pointer,
        return only the part identified by the json pointer.

        :param file_id: the id of the included file, might contain #xpointer
        :return:    loaded json
        """
        if '#' in location:
            file_id, json_pointer_or_id = location.rsplit('#', 1)
        else:
            file_id = location
            json_pointer_or_id = None

        if not file_id or file_id == '.':
            ret = self.schema
        elif file_id.startswith('.'):
            # relative include
            ret = self._load(self.abs_path.parent / file_id)
        else:
            if file_id not in self.included_schemas:
                raise IncludedFileNotFoundException(f'Included file {file_id} not found in includes')

            ret = self.included_schemas[file_id](self)

        if json_pointer_or_id:
            if json_pointer_or_id.startswith('/'):
                ret = resolve_pointer(ret, json_pointer_or_id)
            else:
                ret = resolve_id(ret, json_pointer_or_id)
                if not ret:
                    raise IncludedFileNotFoundException(f'Element with id {json_pointer_or_id} not found in {file_id}')

        ret = copy.deepcopy(ret)
        ret.pop('$id', None)
        ret['oarepo:included-from'] = location
        return ret

    def _resolve_references(self, element, stack):
        if isinstance(element, dict):
            if self.OAREPO_USE in element:
                included_name = element.pop(self.OAREPO_USE)
                if not isinstance(included_name, list):
                    included_name = [included_name]
                for name in included_name:
                    included_data = self._load_included_file(name)
                    deepmerge(element, included_data, [])
                return self._resolve_references(element, stack)
            for k, v in element.items():
                self._resolve_references(v, stack + [k])
        elif isinstance(element, list):
            for v in element:
                self._resolve_references(v, stack)

    @property
    def abs_path(self):
        return Path(self.file_path).absolute()


def resolve_id(json, element_id):
    if isinstance(json, dict):
        if '$id' in json and json['$id'] == element_id:
            return json
        continue_with = json.values()
    elif isinstance(json, (tuple, list)):
        continue_with = json
    else:
        return None
    for k in continue_with:
        ret = resolve_id(k, element_id)
        if ret is not None:
            return ret



import os
from pathlib import Path
from typing import Dict

from . import ModelPreprocessor
from ..schema import ModelSchema


class DefaultValuesModelPreprocessor(ModelPreprocessor):
    TYPE = 'default'

    def transform(self, schema: ModelSchema, settings: Dict):
        self.set(settings, 'package',
                 lambda: os.path.basename(schema.schema.get('output-directory', os.getcwd())).replace('-', '_'))

        self.set(settings, 'processing-order', lambda: ['settings', '*', 'model'])

        self.set(settings, 'package-base', lambda: settings.package.rsplit('.', maxsplit=1)[-1])

        self.set(settings, 'package-base-upper', lambda: settings.package_base.upper())

        self.set(settings, 'kebap-package', lambda: settings.package.replace('_', '-'))

        @self.set(settings, 'package-path')
        def c():
            package_path = settings.package.split('.')
            return Path(package_path[0]).joinpath(*package_path[1:])

        self.set(settings, 'schema-version', lambda: schema.schema.get('version', '1.0.0'))

        self.set(settings, 'schema-name', lambda: f"{settings.kebap_package}-{settings.schema_version}.json")

        self.set(
            settings, 'schema-file', lambda: os.path.join(
                settings.package_path,
                'records',
                'jsonschemas',
                settings.schema_name
            )
        )

        self.set(
            settings, 'mapping-package', lambda: f'{settings.package}.records.mappings'
        )

        self.set(
            settings, 'jsonschemas-package', lambda: f'{settings.package}.records.jsonschemas'
        )

        self.set(
            settings, 'mapping-file', lambda: os.path.join(
                settings.package_path,
                'records',
                'mappings',
                'v7',
                settings.package_base,
                settings.schema_name
            )
        )

        self.set(settings, 'schema-server', lambda: 'http://localhost/schemas/')

        self.set(settings, 'index-name',
                 lambda: settings.package_base + '-' + os.path.basename(settings.mapping_file).replace('.json', ''))

        self.set(settings, 'collection-url', lambda: f'/{settings.package_base}/')



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput
from ..utils.verbose import log


class PoetryBuilder(OutputBuilder):
    TYPE = 'poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        output.setdefault("tool.poetry",
                          "name", self.settings.package_base.replace('_', '-'),
                          "version", "0.0.1",
                          "description", f"A sample application for {self.settings.package}",
                          "authors", [])

        output.setdefault("build-system",
                          "requires", ['poetry-core>=1.0.0'],
                          "build-backend", "poetry.core.masonry.api")

        output.setdefault("tool.poetry.dependencies", "python", "^3.9")

        if 'runtime-dependencies' in self.schema.schema:
            output.table("tool.poetry.dependencies")
            for dep, value in self.schema.schema.runtime_dependencies.items():
                if isinstance(value, str):
                    value = {'version': value}
                output.setdefault("tool.poetry.dependencies." + dep, *sum((list(x) for x in value.items()), []))

        if 'dev-dependencies' in self.schema.schema:
            output.table("tool.poetry.dev-dependencies")
            for dep, value in self.schema.schema.dev_dependencies.items():
                if isinstance(value, str):
                    value = {'version': value}
                output.setdefault("tool.poetry.dev-dependencies." + dep, *sum((list(x) for x in value.items()), []))

        if output.created:
            log(log.INFO, f"""To install the sample app, run
    poetry install -E sample-app            
            """)



from .invenio_base import InvenioBaseClassPythonBuilder

class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_search'
    class_config = 'record-search-options-class'
    template = 'record-search-options'



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_resource'
    class_config = 'record-resource-class'
    template = 'record-resource'



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioRecordJSONSchemasPoetryBuilder(OutputBuilder):
    TYPE = 'invenio_record_jsonschemas_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        output.set(
            "tool.poetry.plugins.'invenio_jsonschemas.schemas'",
            self.settings.python.record_jsonschemas_poetry,
            self.settings.jsonschemas_package
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_service'
    class_config = 'record-service-class'
    template = 'record-service'



from pathlib import Path

from oarepo_model_builder.stack import ModelBuilderStack
from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules
from ..utils.schema import is_schema_element
from ..utils.verbose import log


class MappingBuilder(JSONBaseBuilder):
    TYPE = 'mapping'
    output_file_type = 'mapping'
    output_file_name = 'mapping-file'
    parent_module_root_name = 'mappings'

    @process('/model/**', condition=lambda current, stack: is_schema_element(stack))
    def enter_model_element(self, stack: ModelBuilderStack):
        # ignore schema leaves different than "type" - for example, minLength, ...
        # that should not be present in mapping
        if (
                stack.top_type in (stack.LIST, stack.DICT) or

                # ignore top-level type=object as it is not allowed
                # in elasticsearch mapping (it is object automatically)
                stack.level > 3 and stack.top.key == 'type'
        ):

            self.model_element_enter(stack)

            # process children
            yield

            data = stack.top.data
            if isinstance(data, dict) and 'oarepo:mapping' in data:
                mapping = self.call_components('before_merge_mapping', data['oarepo:mapping'], stack=stack)
                self.output.merge_mapping(mapping)

            self.model_element_leave(stack)

    def on_enter_model(self, output_name, stack: ModelBuilderStack):
        ensure_parent_modules(self.builder, Path(output_name),
                              ends_at=self.parent_module_root_name)
        self.output.merge(
            self.settings.elasticsearch.templates[self.settings.elasticsearch.version]
        )
        self.output.enter('mappings', {})

    def finish(self):
        super().finish()

        log(log.INFO, f"""
    invenio index init --force
            """)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record'
    class_config = 'record-class'
    template = 'record'



from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.model_preprocessors import ModelPreprocessor


class ElasticsearchModelPreprocessor(ModelPreprocessor):
    TYPE = 'elasticsearch'

    def transform(self, schema, settings):
        deepmerge(
            settings,
            {
                'elasticsearch': {
                    'version': 'v7',
                    'keyword-ignore-above': 50,
                    'templates': {
                        'v7': {
                            'mappings': {
                                "properties": {
                                }
                            }
                        }
                    }
                }
            }
        )
